// HashString.h

#ifndef _HASH_STRING_H_
#define _HASH_STRING_H_

#include "type.h"
#include "hash_h.h"
#include "allocator2.h"
#include "allocator_id.h"

#pragma warning ( disable : 4244 )
#pragma warning ( disable : 4267 )

//------------------------------------------------------------------------------


//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
// HashString
//
// hash string in memory:
//
//       bytes
// ptr_ -> 4    - reference counter
//         4    - unsigned hash code
//         4    - [string length (not aligned)]
//         N    - ASCIIZ string
//         NA-N - ( NA = aligned ( N ) )
//------------------------------------------------------------------------------

namespace HashString
{
//------------------------------------------------------------------------------
// HashStringAllocator
//------------------------------------------------------------------------------


class HashStringAllocator
{
public:
	static inline void * alloc ( unsigned size )
	{
		if ( size <= 16 )
			return FixedSmallObjAllocator<16,HashStringAllocatorId,OwnPageAllocator>::allocate();
		else if ( size <= 32 )
			return FixedSmallObjAllocator<32,HashStringAllocatorId,OwnPageAllocator>::allocate();
		else if ( size <= 64 )
			return FixedSmallObjAllocator<64,HashStringAllocatorId,OwnPageAllocator>::allocate();
		else if ( size <= 128 )
			return FixedSmallObjAllocator<128,HashStringAllocatorId,OwnPageAllocator>::allocate();
		else if ( size <= 256 )
			return FixedSmallObjAllocator<256,HashStringAllocatorId,OwnPageAllocator>::allocate();
		else
		{
			ASSERT(0);
			return new byte[size];
		}
	}

	static void free ( void * address, unsigned size ) // size must be same as in alloc() call
	{
		if ( size <= 16 )
			FixedSmallObjAllocator<16,HashStringAllocatorId,OwnPageAllocator>::deallocate ( address );
		else if ( size <= 32 )
			FixedSmallObjAllocator<32,HashStringAllocatorId,OwnPageAllocator>::deallocate ( address );
		else if ( size <= 64 )
			FixedSmallObjAllocator<64,HashStringAllocatorId,OwnPageAllocator>::deallocate ( address );
		else if ( size <= 128 )
			FixedSmallObjAllocator<128,HashStringAllocatorId,OwnPageAllocator>::deallocate ( address );
		else if ( size <= 256 )
			FixedSmallObjAllocator<256,HashStringAllocatorId,OwnPageAllocator>::deallocate ( address );
		else
		{
			ASSERT(0);
			delete[] address;
		}
	}

	static const int align = 8;
};


//------------------------------------------------------------------------------
//
//#if VERIFY_PHS
//	extern std::set<byte*> phs_set_;
//#endif
//
////------------------------------------------------------------------------------
//
//inline bool use_own_allocator ( unsigned size ) // string size in bytes + '\0'
//{
//	return size <= allocator_max_len_;
//}
//
////------------------------------------------------------------------------------
//
//inline PHashString clone ( const PHashString phs )
//{
//	invariant ( phs );
//
//	unsigned size = getLength ( phs ) + 1;
//	byte * str;
//	
//	if ( use_own_allocator ( size ) )
//	{
//		str = (byte*) HashStringAllocator::alloc ( size + data_size );
//	}
//	else
//	{
//		str = (byte*)new byte[size+data_size];
//	}
//	if ( !str )
//		return PHashString ( 0 );
//
//	memcpy ( (char*) & str[0], (const char*) & phs.address_[0], size + data_size );
//
//#if VERIFY_PHS
////	TRACE2("%p clone %u\n",str,size-1);
//	ASSERT ( phs_set_.find ( str ) == phs_set_.end() );
//	phs_set_.insert ( str );
//#endif
//
//	invariant ( PHashString ( str ) );
//
//	return PHashString ( str );
//}
//
////------------------------------------------------------------------------------
//
//inline PHashString internalCreateHashString
//	( const Char * s, unsigned size, unsigned hash_code )
//{
//	++size; // number of chars + '\0'
//	if ( size > max_len_ )
//		size = max_len_;
//	byte * str;
//
//	if ( use_own_allocator ( size ) )
//	{
//		str = (byte*) HashStringAllocator::alloc ( size + data_size );
//	}
//	else
//	{
//		str = (byte*)new byte[size+data_size];
//	}
//	if ( !str )
//		return PHashString ( 0 );
//
//	*((unsigned *) (&str[0])) = hash_code;
//	*((unsigned *) (&str[4])) = size-1;
//	memcpy ( (char*) & str[data_size], (const char*) s, size - 1 );
//	str[data_size+size-1] = '\0';
//
//#if VERIFY_PHS
////	TRACE2("%p create %u\n",str,size-1);
//	ASSERT ( phs_set_.find ( str ) == phs_set_.end() );
//	phs_set_.insert ( str );
//#endif
//
//	invariant ( PHashString ( str ) );
//
//	return PHashString ( str );
//}
//
//inline PHashString createHashString ( const Char * s )
//{
//	const Char * s_begin = s;
//
//// eval hash code
//
//	hash_code_type hashCode;
//	hashCodeStart ( hashCode );
//	while ( *s )
//		hashCodeAdd ( hashCode, *s++ );
//
//	return internalCreateHashString
//		( s_begin, s - s_begin, hashCode );
//}
//
////------------------------------------------------------------------------------
//
//inline PHashString createHashString
//	( const Char * s_begin, const Char * s_end )
//{
//	const Char * s = s_begin;
//
//// eval hash code
//
//	hash_code_type hashCode;
//	hashCodeStart ( hashCode );
//	while ( s != s_end )
//		hashCodeAdd ( hashCode, *s++ );
//
//	return internalCreateHashString
//		( s_begin, s_end - s_begin, hashCode );
//}
//
////------------------------------------------------------------------------------
//
//inline PHashString createHashString ( const Char * s, unsigned hash_code )
//{
//	return internalCreateHashString
//		( s, strlen ( (const char*) s ), hash_code );
//}
//
////------------------------------------------------------------------------------
//
//inline PHashString createHashString
//	( const Char * s_begin, const Char * s_end, unsigned hash_code )
//{
//	return internalCreateHashString
//		( s_begin, s_end - s_begin, hash_code );
//}
//
////------------------------------------------------------------------------------
//
//inline void deleteHashString ( PHashString phs )
//{
//	invariant ( phs );
//
//#if VERIFY_PHS
////	TRACE1("%p delete\n",phs.address_);
//	std::set<byte*>::iterator it;
//	ASSERT ( ( it = phs_set_.find ( phs.address_ ) ) != phs_set_.end() );
//	phs_set_.erase ( *it );
//#endif
//
//	if ( use_own_allocator ( getLength ( phs ) + 1 ) )
//	{
////		delete [] ((byte*)phs.address_);
//		HashStringAllocator::free ( (void*)phs.address_, getLength ( phs ) + 1 + data_size );
//	}
//	else
//	{
//		delete [] ((byte*)phs.address_);
//	}
//}
//
////------------------------------------------------------------------------------
//
//inline unsigned getHashCode ( const PHashString phs )
//{
//	return *((unsigned*)(&phs.address_[0]));
//}
//
////------------------------------------------------------------------------------
//
//inline unsigned getLength ( const PHashString phs )
//{
//	return *((unsigned*)(&phs.address_[4]));
//}
//
////------------------------------------------------------------------------------
//
//inline const Char * getString ( const PHashString phs )
//{
//	return (const Char*)(&phs.address_[8]);
//}
//
////------------------------------------------------------------------------------
//
//inline bool equal ( const PHashString lhs, const PHashString rhs )
//{
//	return ( getHashCode(lhs) == getHashCode(rhs) )
//		&& ( strcmp ( (const char*)getString ( lhs ),
//		              (const char*)getString ( rhs ) ) == 0 );
//}
//
////------------------------------------------------------------------------------
//
//inline void invariant ( const PHashString phs )
//{
//	if ( !phs.isNull() )
//	{
//		ASSERT ( strlen ( (const char*)getString ( phs ) )
//			== getLength ( phs ) );
//
//		const Char * s = getString ( phs );
//		const Char * s_end = s + getLength ( phs );
//		
//		ASSERT ( *s_end == '\0' );
//		
//		hash_code_type hashCode;
//		hashCodeStart ( hashCode );
//		while ( s != s_end )
//			hashCodeAdd ( hashCode, *s++ );
//
//		ASSERT ( getHashCode ( phs ) == hashCode );
//	}
//	else
//	{
//		ASSERT ( getString ( phs ) == 0 );
//		ASSERT ( getLength ( phs ) == 0 );
//
//		hash_code_type hashCode;
//		hashCodeStart ( hashCode );
//
//		ASSERT ( getHashCode ( phs ) == hashCode );
//	}
//}

//------------------------------------------------------------------------------
// PHashString/WPHashString
//
// HashString with reference count
//
// hash string in memory:
//
//       bytes
// ptr_ -> 4    - reference count
//         4    - unsigned hash code
//         4    - [string length (not aligned)]
//         N    - ASCIIZ string
//         NA-N - ( NA = aligned ( N ) )
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// WPHashString
//
// weak temple pointer to hash string with reference counter
// used to access to PHashString without incrementing and decrementing
// reference counter
//------------------------------------------------------------------------------


class WPHashString
{
	friend class PHashString;

public:
	explicit WPHashString()
		: address_ ( 0 )
	{}

	//WPHashString ( const PHashString & rhs )
	//	: address_ ( rhs.address_ )
	//{}

private:
	WPHashString ( byte * address )
		: address_ ( address )
	{}

private:
	byte * address_;

public:
	inline operator bool() const
	{
		return address_ != 0;
	}

	friend inline bool isNull ( WPHashString wphs );
	friend inline unsigned getHashCode ( WPHashString wphs );
	friend inline unsigned getLength ( WPHashString wphs );
	friend inline bool isVerbatim ( WPHashString wphs );
	friend inline void setVerbatim ( WPHashString wphs, bool verbatim );
	friend inline const Char * getString ( WPHashString wphs );
	friend inline bool equal ( WPHashString lhs, WPHashString rhs );
	friend inline bool empty ( WPHashString wphs );
	friend inline void invariant ( WPHashString wphs );
};


//------------------------------------------------------------------------------
// PHashString
//------------------------------------------------------------------------------


class PHashString
{
	friend class WPHashString;
	friend inline bool isNull ( WPHashString wphs );
	friend inline unsigned getHashCode ( WPHashString wphs );
	friend inline unsigned getLength ( WPHashString wphs );
	friend inline bool isVerbatim ( WPHashString wphs );
	friend inline void setVerbatim ( WPHashString wphs, bool verbatim );
	friend inline const Char * getString ( WPHashString wphs );
	friend inline bool equal ( WPHashString lhs, WPHashString rhs );
	friend inline bool empty ( WPHashString wphs );
	friend inline void invariant ( WPHashString wphs );

private:

	static const int allocator_record_size_max_ = 256;
	static const int ref_count_offset_ = 0;
	static const int hash_code_offset_ = 4;
	static const int length_offset_ = 8;
	static const int string_offset_ = 12;
	static const int data_size_offset_ = 12;
	static const int allocator_max_len_ = allocator_record_size_max_ - data_size_offset_;
	//static const int max_len_ = 65536; // max number of chars + '\0'

//------------------------------------------------------------------------------

	static inline bool use_own_allocator ( unsigned size ) // string size in bytes + '\0'
	{
		return size <= allocator_max_len_;
	}

//------------------------------------------------------------------------------

	inline unsigned getLength() const
	{
		if ( address_ )
		{
			unsigned len = *((unsigned*)(&address_[length_offset_]));
			len &= 0x7fffffffU;
			return len;
		}
		else
			return 0;
	}

//------------------------------------------------------------------------------

	inline bool isVerbatim() const
	{
		if ( address_ )
		{
			unsigned len = *((unsigned*)(&address_[length_offset_]));
			return ( len & 0x80000000U ) != false;
		}
		else
			return 0;
	}

//------------------------------------------------------------------------------

	inline void setVerbatim ( bool verbatim )
	{
		unsigned len = *((unsigned*)(&address_[length_offset_]));
		if ( verbatim )
			len |= 0x80000000U;
		else
			len &= (~0x80000000U);
		*((unsigned*)(&address_[length_offset_])) = len;
	}

//------------------------------------------------------------------------------

	inline void internalCreateHashString
		( const Char * s, unsigned size, unsigned hashCode )
	{
		++size; // number of chars + '\0'
		//if ( size > max_len_ )
		//	size = max_len_;
		byte * str;

		const int alloc_size = size + data_size_offset_;


		if ( use_own_allocator ( size ) )
		{
			str = (byte*) HashStringAllocator::alloc ( alloc_size );
		}
		else
		{
			str = (byte*)new byte[alloc_size];
		}


		if ( !str )
			address_ = 0;

		*((unsigned *) (&str[ref_count_offset_])) = 1;
		*((unsigned *) (&str[hash_code_offset_])) = hashCode;
		*((unsigned *) (&str[length_offset_])) = size-1;
		memcpy ( (char*) & str[string_offset_], (const char*) s, size - 1 );
		str[alloc_size-1] = '\0';

	//#if VERIFY_PHS
	//	//	TRACE2("%p create %u\n",str,size-1);
	//	ASSERT ( phs_set_.find ( str ) == phs_set_.end() );
	//	phs_set_.insert ( str );
	//#endif

	//	invariant ( PHashString ( str ) );

		address_ = str;
	}

//------------------------------------------------------------------------------

	inline void createHashString ( const Char * s )
	{
		const Char * s_begin = s;

	// eval hash code

		hash_code_type hashCode;
		hashCodeStart ( hashCode );
		while ( *s )
			hashCodeAdd ( hashCode, *s++ );

		internalCreateHashString ( s_begin, s - s_begin, hashCode );
	}

//------------------------------------------------------------------------------

	inline void createHashString
		( const Char * s_begin, const Char * s_end )
	{
		const Char * s = s_begin;

	// eval hash code

		hash_code_type hashCode;
		hashCodeStart ( hashCode );
		while ( s != s_end )
			hashCodeAdd ( hashCode, *s++ );

		internalCreateHashString ( s_begin, s_end - s_begin, hashCode );
	}

//------------------------------------------------------------------------------

	inline void createHashString ( const Char * s, unsigned hash_code )
	{
		internalCreateHashString
			( s, strlen ( (const char*) s ), hash_code );
	}

//------------------------------------------------------------------------------

	inline void createHashString
		( const Char * s_begin, const Char * s_end, unsigned hash_code )
	{
		return internalCreateHashString
			( s_begin, s_end - s_begin, hash_code );
	}

//------------------------------------------------------------------------------

	inline void deleteHashString()
	{
		//invariant ( phs );

	//#if VERIFY_PHS
	//	//	TRACE1("%p delete\n",phs.address_);
	//	std::set<byte*>::iterator it;
	//	ASSERT ( ( it = phs_set_.find ( phs.address_ ) ) != phs_set_.end() );
	//	phs_set_.erase ( *it );
	//#endif


		if ( use_own_allocator ( getLength() + 1 ) )
		{
			HashStringAllocator::free ( address_, getLength() + 1 + data_size_offset_ );
		}
		else
		{
			delete [] address_;
		}

	}

//------------------------------------------------------------------------------

	inline void Lock()
	{
		if ( address_ )
			++((unsigned&)address_[ref_count_offset_]);
	}

//------------------------------------------------------------------------------

	inline void Unlock()
	{
		if ( address_ )
		{
			if ( !--((unsigned&)address_[ref_count_offset_]) )
				deleteHashString();
		}
	}

//------------------------------------------------------------------------------

public:
	PHashString() : address_(0) {}

	~PHashString()
	{
		Unlock();
	}

	inline operator const WPHashString() const
	{
		return WPHashString ( address_ );
	}

	inline operator WPHashString()
	{
		return WPHashString ( address_ );
	}

	inline bool isNull() const
	{
		return address_ == 0;
	}

	inline operator bool() const
	{
		return address_ != 0;
	}

	inline PHashString ( const PHashString & rhs )
		: address_(rhs.address_)
	{
		Lock();
	}

	inline PHashString ( const WPHashString & rhs )
		: address_(rhs.address_)
	{
		Lock();
	}

	inline PHashString ( const Char * s )
	{
		if ( s )
			createHashString ( s );
		else
			address_ = 0;
	}

	inline PHashString ( const Char * s_begin, const Char * s_end )
	{
		createHashString ( s_begin, s_end );
	}

	inline PHashString ( const Char * s, unsigned hashCode )
	{
		createHashString ( s, hashCode );
	}

	inline PHashString ( const Char * s_begin, const Char * s_end, unsigned hashCode )
	{
		createHashString ( s_begin, s_end, hashCode );
	}

//------------------------------------------------------------------------------

	inline PHashString & operator = ( const PHashString & rhs )
	{
		if ( this == & rhs )
			return *this;

		Unlock();

		address_ = rhs.address_;
		
		Lock();

		return *this;
	}

	inline PHashString & operator = ( const WPHashString & rhs )
	{
		Unlock();

		address_ = rhs.address_;

		Lock();

		return *this;
	}

//------------------------------------------------------------------------------

private:
	inline bool operator == ( const PHashString & rhs ) const
	{
		return address_ == rhs.address_;
	}

private:
	byte * address_;
};

//------------------------------------------------------------------------------

inline bool isNull ( WPHashString wphs )
{
	return wphs.address_ == 0;
}

//------------------------------------------------------------------------------

inline unsigned getHashCode ( WPHashString wphs )
{
	if ( wphs.address_ )
		return *((unsigned*)(&wphs.address_[PHashString::hash_code_offset_]));
	else
		return 0;
}

//------------------------------------------------------------------------------

inline unsigned getLength ( WPHashString wphs )
{
	if ( wphs.address_ )
	{
		unsigned len = *((unsigned*)(&wphs.address_[PHashString::length_offset_]));
		len &= 0x7fffffffU;
		return len;
	}
	else
		return 0;
}

//------------------------------------------------------------------------------

inline bool isVerbatim ( WPHashString wphs )
{
	if ( wphs.address_ )
	{
		unsigned len = *((unsigned*)(&wphs.address_[PHashString::length_offset_]));
		return ( len & 0x80000000U ) != 0;
	}
	else
		return 0;
}

//------------------------------------------------------------------------------

inline void setVerbatim ( WPHashString wphs, bool verbatim )
{
	unsigned len = *((unsigned*)(&wphs.address_[PHashString::length_offset_]));
	if ( verbatim )
		len |= 0x80000000U;
	else
		len &= (~0x80000000U);
	*((unsigned*)(&wphs.address_[PHashString::length_offset_])) = len;
}

//------------------------------------------------------------------------------

inline const Char * getString ( WPHashString wphs )
{
	if ( wphs.address_ )
		return ((const Char*)(&wphs.address_[PHashString::string_offset_]));
	else
		return 0;
}

//------------------------------------------------------------------------------

inline bool equal ( WPHashString lhs, WPHashString rhs )
{
	if ( isNull ( lhs ) || isNull ( rhs ) )
		return isNull ( lhs ) && isNull ( rhs );
	return ( getHashCode ( lhs ) == getHashCode ( rhs )
		//&& getLength ( lhs ) == getLength ( rhs )
		&& strcmp ( (const char*) getString ( lhs ),
			          (const char*) getString ( rhs )//,
								/*getLength ( lhs )*/ ) == 0 );
}

//------------------------------------------------------------------------------

inline bool empty ( WPHashString wphs )
{
	return isNull ( wphs ) || ( getLength ( wphs ) == 0 );
}

//------------------------------------------------------------------------------

inline void invariant ( const WPHashString wphs )
{
	if ( !isNull ( wphs ) )
	{
		ASSERT ( strlen ( (const char*) getString ( wphs ) )
			== getLength ( wphs ) );

		const Char * s = getString ( wphs );
		const Char * s_end = s + getLength ( wphs );
		
		ASSERT ( *s_end == '\0' );
		
		hash_code_type hashCode;
		hashCodeStart ( hashCode );
		while ( s != s_end )
			hashCodeAdd ( hashCode, *s++ );

		ASSERT ( getHashCode ( wphs ) == hashCode );
	}
	else
	{
		ASSERT ( getString ( wphs ) == 0 );
		ASSERT ( getLength ( wphs ) == 0 );

		hash_code_type hashCode;
		hashCodeStart ( hashCode );

		ASSERT ( getHashCode ( wphs ) == hashCode );
	}
}

} // namespace HashString

//------------------------------------------------------------------------------


//------------------------------------------------------------------------------

#endif // _HASH_STRING_H_
