// ICsProject.h

#pragma once

#ifndef _I_CS_PARSER_H_
#define _I_CS_PARSER_H_

class ICsFile;
class IAssembly;
class PerfomanceStatistics;
struct CsEntityNamespace;
struct CsEntityMethod;

//------------------------------------------------------------------------------
// ICsProject
//------------------------------------------------------------------------------

class ICsProject
{
public:
	static ICsProject * create();
	virtual ~ICsProject();

// add/remove files
	virtual void attachFile ( ICsFile * file ) =0;
	virtual void attachFiles ( int n_files, ICsFile ** files ) =0;
	virtual void getFiles ( int & n_files, ICsFile **& files ) const =0;
	virtual void detachFile ( ICsFile * file ) =0;
	virtual void detachFiles ( int n_files, ICsFile ** files ) =0;
	virtual void detachAllFiles() =0;

// add/remove assembly
	virtual void attachAssembly ( IAssembly * assembly ) =0;
	virtual void attachAssemblies ( int n_assemblies, IAssembly ** assemblies ) =0;
	virtual void getAssemblies ( int & n_assemblies, IAssembly **& assemblies ) const =0;
	virtual void detachAssembly ( IAssembly * assembly ) =0;
	virtual void detachAssemblies ( int n_assemblies, IAssembly ** assemblies ) =0;
	virtual void detachAllAssemblies() =0;

// add/remove macro
	virtual void addMacro ( const char * macro ) =0;
	virtual void addMacros ( int n_macros, const char ** macros ) =0;
	virtual void getMacros ( int & n_macros, const char **& macros ) const =0;
	virtual void getMacros_free ( int n_macros, const char ** macros ) const =0;
	virtual bool isMacroSet ( const char * macros ) const =0;
	virtual void removeMacro ( const char * macros ) =0;
	virtual void removeMacros ( int n_macros, const char ** macros ) =0;
	virtual void removeAllMacros() =0;

// generate compilation units and entity model
	virtual void parse() =0;


// reparse entities
	virtual void reparseEntities() =0;

// reparse file (AST only)
	virtual void reparseFiles ( int n_files, ICsFile ** files ) =0;

// reparse local entities
	virtual void reparseLocalEntities ( CsEntityMethod * method ) =0;


// delete all generated compilation units and entity model
	virtual void clear() =0;

// get Display
	virtual CsEntityNamespace * getGlobalNamespace() const =0;

// perfomance statistics
	virtual const PerfomanceStatistics & getPerfomanceStatistics() const =0;

// Build Entity Model flag (passes 2,3,4 and metadata)
	virtual void setBuildEntityModel ( bool flag ) =0;
	virtual bool getBuildEntityModel() const =0;

protected:
	ICsProject();

private:
	ICsProject ( const ICsProject & rhs );
	ICsProject & operator= ( const ICsProject & rhs );
};

//------------------------------------------------------------------------------

#endif // _I_CS_PARSER_H_
