// CsFormatter.cpp

#include "stdafx.h"
#include "CsFormatter.h"

#include <limits>


#undef min
#undef max

//------------------------------------------------------------------------------
// CsFormatter
//------------------------------------------------------------------------------

//size_t CsFormatter::st_index = 0;

//------------------------------------------------------------------------------

CsFormatter::CsFormatter()
	: storeNewlines_ ( true )
	, tab_token_buffer_ ( 0 )
	, tab_token_buffer_size_ ( 0 )
	, ws_token_buffer_ ( 0 )
	, ws_token_buffer_size_ ( 0 )
{
	setDefault();
}

//------------------------------------------------------------------------------

CsFormatter::~CsFormatter()
{
	delete[] tab_token_buffer_;
	delete[] ws_token_buffer_;
}

//------------------------------------------------------------------------------

void CsFormatter::setDefault()
{
	compound_identifier_before_after.clear();
	compound_identifier_around_dot.clear();
	keyword_around.clear();
	keyword_around.push_back ( CsToken ( CsTokenType ( tkF_WS ), 1, LineInfo() ) );
	operator_around.clear();
	operator_around.push_back ( CsToken ( CsTokenType ( tkF_WS ), 1, LineInfo() ) );
	end_of_line_type_ = windows_line;
	generation_option_ = go_generate_text;
}

//------------------------------------------------------------------------------

void CsFormatter::generate ( CsNode * node )
{
	//{
	//	std::string  s ( node->getFilePath() );
	//	XASSERT ( s.find ( "CharScanner.cs" ) == -1 );
	//}

	tokens_.clear();
	generated_text_.clear();
	n_tab = 0;
//#ifdef _DEBUG
//	bCheckPushBack = true;
//#endif // _DEBUG
	addNewlineKill0();
	node->callback ( this );
	addNewlineKill1();
//#ifdef _DEBUG
//	bCheckPushBack = false;
//#endif // _DEBUG
	
	MediumFastDeque<CsToken>::iterator it = tokens_.begin();
	MediumFastDeque<CsToken>::iterator it_end = tokens_.end();
	while ( it != it_end )
	{
//#ifdef _DEBUG
//		size_t index = std::distance ( tokens_.begin(), it );
//#endif // _DEBUG

		unsigned n_ws_min = 0;
		unsigned n_ws_max = std::numeric_limits<unsigned>::max();
		unsigned n_newline_min = 0;
		unsigned n_newline_max = std::numeric_limits<unsigned>::max();

		while ( it != it_end
			&& ( it->type == tkF_WS
				|| it->type == tkF_WS_KILL
				|| it->type == tkF_NEWLINE
				|| it->type == tkF_NEWLINE_KILL
				|| it->type == tkF_INC_TAB
				|| it->type == tkF_DEC_TAB ) )
		{
			// ws

ws_loop:
			while ( it != it_end
				&& ( it->type == tkF_WS
					|| it->type == tkF_WS_KILL ) )
			{
				if ( it->type == tkF_WS )
				{
					n_ws_min = std::max ( n_ws_min, (unsigned)it->number );
					++it;
					goto ws_loop;
				}
				if ( it->type == tkF_WS_KILL )
				{
					n_ws_max = std::min ( n_ws_max, (unsigned)it->number );
					++it;
					goto ws_loop;
				}
			}
			
			// newline

newline_loop:
			if ( it != it_end
				&& ( it->type == tkF_NEWLINE
					|| it->type == tkF_NEWLINE_KILL ) )
			{
				if ( it->type == tkF_NEWLINE )
				{
					n_newline_min = std::max ( n_newline_min, (unsigned)it->number );
					//n_ws_max = 0;
					++it;
					goto newline_loop;
				}
				if ( it->type == tkF_NEWLINE_KILL )
				{
					n_newline_max = std::min ( n_newline_max, (unsigned)it->number );
					++it;
					goto newline_loop;
				}
			}

			if ( it != it_end
				&& it->type == tkF_INC_TAB )
			{
				++n_tab;
				++it;
			}
			if ( it != it_end
				&& it->type == tkF_DEC_TAB )
			{
				--n_tab;
				++it;
			}
		}

		unsigned n_newline = std::min ( n_newline_min, n_newline_max );
		if ( n_newline )
			n_ws_max = 0;
		for ( unsigned i = 0; i != n_newline; ++i )
		{
			if ( generation_option_ == go_generate_text )
			{
				switch ( end_of_line_type_ )
				{
					case windows_line:
					default:
						generated_text_.push_back ( '\r' );
						generated_text_.push_back ( '\n' );
						break;
					case unix_line:
						generated_text_.push_back ( '\n' );
						break;
					case mac_line:
						generated_text_.push_back ( '\r' );
						break;
				}
			}
			else if ( generation_option_ == go_generate_tokens )
			{
				generated_tokens_.push_back
					( CsToken ( tkNEW_LINE, end_of_line_type_, LineInfo() ) );
			}
			else
				ASSERT(0);
			if ( generation_option_ == go_generate_text )
			{
				for ( unsigned i = 0; i != n_tab; ++i )
				{
					generated_text_.push_back ( '\t' );
				}
			}
			else if ( generation_option_ == go_generate_text )
			{
				if ( n_tab > tab_token_buffer_size_ )
				{
					while ( n_tab > tab_token_buffer_size_ )
					{
						if ( tab_token_buffer_size_ == 0 )
							tab_token_buffer_size_ = tab_token_buffer_size_start_;
						else
							tab_token_buffer_size_ *= 2;
					}
					delete[] tab_token_buffer_;
					tab_token_buffer_ = new char[tab_token_buffer_size_];
					memset ( tab_token_buffer_, '\t', tab_token_buffer_size_ );
				}
				generated_tokens_.push_back
					( CsToken ( tkWS, (const Char*)tab_token_buffer_, (const Char*)tab_token_buffer_ + n_tab, LineInfo() ) );
			}
			else
				ASSERT(0);
		}

		unsigned n_ws = std::min ( n_ws_min, n_ws_max );
		if ( generation_option_ == go_generate_text )
		{
			for ( unsigned i = 0; i != n_ws; ++i )
			{
				generated_text_.push_back ( ' ' );
			}
		}
		else if ( generation_option_ == go_generate_text )
		{
			if ( n_ws > ws_token_buffer_size_ )
			{
				while ( n_ws > ws_token_buffer_size_ )
				{
					if ( ws_token_buffer_size_ == 0 )
						ws_token_buffer_size_ = ws_token_buffer_size_start_;
					else
						ws_token_buffer_size_ *= 2;
				}
				delete[] ws_token_buffer_;
				ws_token_buffer_ = new char[ws_token_buffer_size_];
				memset ( ws_token_buffer_, ' ', ws_token_buffer_size_ );
			}
			generated_tokens_.push_back
				( CsToken ( tkWS, (const Char*)ws_token_buffer_, (const Char*)ws_token_buffer_ + n_ws, LineInfo() ) );
		}
		else
			ASSERT(0);

		if ( it != it_end )
		{
			if ( generation_option_ == go_generate_text )
			{
				const char * text = (const char*)it->getText();
				int len = strlen ( text );
				generated_text_.push_back ( text, text + len );
			}
			else if ( generation_option_ == go_generate_tokens )
			{
				generated_tokens_.push_back ( *it );
			}
			else
				ASSERT(0);
			++it;
		}
	}
}

//------------------------------------------------------------------------------
// CsCompoundIdentifier
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsCompoundIdentifier & compound_identifier )
{
	addTokens ( compound_identifier_before_after );
	value_single_list1<CsIdentifier>::iterator it
		= compound_identifier.begin();
	value_single_list1<CsIdentifier>::iterator it_end
		= compound_identifier.end();
	bool bFirst = true;
	while ( it != it_end )
	{
		if ( !bFirst )
		{
			addTokens ( compound_identifier_around_dot );
			addToken ( tkDOT );
			addTokens ( compound_identifier_around_dot );
		}
		addToken ( it->identifier );
		bFirst = false;
		++it;
	}
	addTokens ( compound_identifier_before_after );
}

////------------------------------------------------------------------------------
//// CsUsingDirective
////------------------------------------------------------------------------------
//
//#if CS_TRANSFORMATION_PARSER
//void CsFormatter::callback ( CsUsingDirective * node )
//{
//	addNewline();
//	addKeyword ( tkUSING );
//	addWS();
//	if ( node->bUsingAliasDirective )
//	{
//		addToken ( node->identifier );
//		addOperator ( tkASSIGN );
//		addWS();
//		callback ( node->namespace_or_type_name );
//	}
//	else
//	{
//		callback ( node->namespace_or_type_name );
//	}
//	addSemicolon();
//	addNewline();
//}
//#endif // CS_TRANSFORMATION_PARSER

//------------------------------------------------------------------------------
// CsUsingAliasDirective
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsUsingAliasDirective * node )
{
	addNewline();
	addKeyword ( tkUSING );
	addWS();
	addToken ( node->identifier );
	addOperator ( tkASSIGN );
	addWS();
	callback ( node->namespace_or_type_name );
	addSemicolon();
	addNewline();
}

//------------------------------------------------------------------------------
// CsUsingNamespaceDirective
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsUsingNamespaceDirective * node )
{
	addNewline();
	addKeyword ( tkUSING );
	addWS();
	callback ( node->namespace_or_type_name );
	addSemicolon();
	addNewline();
}

//------------------------------------------------------------------------------
// CsEmbeddedType
//------------------------------------------------------------------------------

//#if CS_TRANSFORMATION_PARSER
//void CsFormatter::callback ( CsEmbeddedType & node )
//{
//	addKeyword ( node.type );
//}
//#endif // CS_TRANSFORMATION_PARSER

//------------------------------------------------------------------------------
// CsTypeSpecifier
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsTypeSpecifier & node )
{
	switch ( node.specifier_type )
	{
		case tsArrayPostfix:
			{
				addToken ( tkLBRACKET );
				for ( unsigned i = 1; i < node.rank; ++i )
				{
					addToken ( tkCOMMA );
				}
				addToken ( tkRBRACKET );
			}
			break;
		case tsPointer:
			{
				addToken ( tkSTAR );
			}
			break;
	}
}

//------------------------------------------------------------------------------
// CsTypeSpecifiers
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsTypeSpecifiers * node )
{
	value_single_list1_with_delete_value<CsTypeSpecifier>::iterator it
		= node->specifiers.begin();
	value_single_list1_with_delete_value<CsTypeSpecifier>::iterator it_end
		= node->specifiers.end();
	while ( it != it_end )
	{
		callback ( *it );
		++it;
	}
}

//------------------------------------------------------------------------------
// CsTypeRef
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsTypeRef * node )
{
	addWS();
	if ( node->isEntityType() )
		callback ( node->compound_identifier );
	else
		addToken ( node->embedded_type );
	if ( node->specifiers )
	{
		addWSKill();
		callback ( node->specifiers );
	}
	addWS();
}

//------------------------------------------------------------------------------
// CsSimpleName
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsSimpleName * node )
{
	addToken ( node->identifier );
}

//------------------------------------------------------------------------------
// CsLiteral
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsLiteral * node )
{
	switch ( node->type )
	{
		case tkLITERAL:
			addToken ( CsToken ( tkLITERAL, node->string, node->literal_type, LineInfo() ) );
			break;
		default:
			addToken ( node->type );
	}
}

//------------------------------------------------------------------------------
// CsEmbeddedTypeMemberAccess
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsEmbeddedTypeMemberAccess * node )
{
	addToken ( node->embedded_type );
	addToken ( tkDOT );
	addToken ( node->member );
}

//------------------------------------------------------------------------------
// CsThisAccess
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsThisAccess * node )
{
	addToken ( tkTHIS );
}

//------------------------------------------------------------------------------
// CsBaseMemberAccess
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsBaseMemberAccess * node )
{
	addToken ( tkBASE );
	addToken ( tkDOT );
	addToken ( node->member );
}

//------------------------------------------------------------------------------
// CsBaseIndexerAccess
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsBaseIndexerAccess * node )
{
	addToken ( tkBASE );
	addToken ( tkLBRACKET );
	if ( node->expression_list )
		callback ( node->expression_list );
	addToken ( tkRBRACKET );
}

//------------------------------------------------------------------------------
// CsTypeofExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsTypeofExpression * node )
{
	addKeyword ( tkTYPEOF );
	addToken ( tkLPAREN );
	if ( node->type )
		callback ( node->type );
	addToken ( tkRPAREN );
}

//------------------------------------------------------------------------------
// CsCheckedExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsCheckedExpression * node )
{
	addKeyword ( tkCHECKED );
	addToken ( tkLPAREN );
	if ( node->expression )
		node->expression->callback ( this );
	addToken ( tkRPAREN );
}

//------------------------------------------------------------------------------
// CsUncheckedExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsUncheckedExpression * node )
{
	addKeyword ( tkUNCHECKED );
	addToken ( tkLPAREN );
	if ( node->expression )
		node->expression->callback ( this );
	addToken ( tkRPAREN );
}

//------------------------------------------------------------------------------
// CsSizeofExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsSizeofExpression * node )
{
	addKeyword ( tkSIZEOF );
	addToken ( tkLPAREN );
	if ( node->type )
		callback ( node->type );
	addToken ( tkRPAREN );
}

//------------------------------------------------------------------------------
// CsParenthesizedExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsParenthesizedExpression * node )
{
	addToken ( tkLPAREN );
	addWS();
	if ( node->expression )
		node->expression->callback ( this );
	addWS();
	addToken ( tkRPAREN );
}

//------------------------------------------------------------------------------
// CsNewObjectExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsNewObjectExpression * node )
{
	addKeyword ( tkNEW );
	if ( node->type )
		callback ( node->type );
	addToken ( tkLPAREN );
	if ( node->argument_list )
	{
		addWS();
		callback ( node->argument_list );
		addWS();
	}
	addToken ( tkRPAREN );
}

//------------------------------------------------------------------------------
// CsNewArrayExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsNewArrayExpression * node )
{
	addKeyword ( tkNEW );
	if ( node->type )
		callback ( node->type );
	addWSKill();
	if ( node->expressions )
	{
		addToken ( tkLBRACKET );
		callback ( node->expressions );
		addToken ( tkRBRACKET );
	}
	if ( node->specifiers )
	{
		addWSKill();
		value_single_list1_with_delete_value<CsTypeSpecifier>::iterator it
			= node->specifiers->begin();
		value_single_list1_with_delete_value<CsTypeSpecifier>::iterator it_end
			= node->specifiers->end();
		while ( it != it_end )
		{
			callback ( *it );
			++it;
		}
	}
	addWS();
	if ( node->initializer )
		callback ( node->initializer );
}

//------------------------------------------------------------------------------
// CsInvocationExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsInvocationExpression * node )
{
	if ( node->expression )
		node->expression->callback ( this );
	addWS();
	addToken ( tkLPAREN );
	if ( node->argument_list
		&& !node->argument_list->list.empty() )
	{
		addWS();
		callback ( node->argument_list );
		addWS();
	}
	addToken ( tkRPAREN );
}

//------------------------------------------------------------------------------
// CsElementAccess
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsElementAccess * node )
{
	if ( node->expression )
		node->expression->callback ( this );
	addToken ( tkLBRACKET );
	if ( node->expressions )
		callback ( node->expressions );
	addToken ( tkRBRACKET );
}

//------------------------------------------------------------------------------
// CsMemberAccess
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsMemberAccess * node )
{
	if ( node->expression )
		node->expression->callback ( this );
	addToken ( tkDOT );
	addToken ( node->member );
}

//------------------------------------------------------------------------------
// CsPointerMemberAccess
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsPointerMemberAccess * node )
{
	if ( node->expression )
		node->expression->callback ( this );
	addToken ( tkARROW );
	addToken ( node->member );
}

//------------------------------------------------------------------------------
// CsPostIncrementDecrementExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsPostIncrementDecrementExpression * node )
{
	if ( node->expression )
		node->expression->callback ( this );
	addToken ( node->oper );
}

//------------------------------------------------------------------------------
// CsUnaryExpression
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// CsBinaryExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsBinaryExpression * node )
{
	if ( node->lhs )
		node->lhs->callback ( this );
	addWS();
	addToken ( node->oper );
	addWS();
	if ( node->rhs )
		node->rhs->callback ( this );
}

//------------------------------------------------------------------------------
// CsAsIsExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsAsIsExpression * node )
{
	if ( node->expression )
		node->expression->callback ( this );
	addKeyword ( node->oper );
	if ( node->type )
		callback ( node->type );
}

//------------------------------------------------------------------------------
// CsConditionalExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsConditionalExpression * node )
{
	if ( node->condition )
		node->condition->callback ( this );
	addOperator ( tkQUESTION );
	if ( node->true_expression )
		node->true_expression->callback ( this );
	addOperator ( tkCOLON );
	if ( node->false_expression )
		node->false_expression->callback ( this );
}

//------------------------------------------------------------------------------
// CsOperatorUnaryExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsOperatorUnaryExpression * node )
{
	addToken ( node->oper );
	if ( node->unary_expression )
		node->unary_expression->callback ( this );
}

//------------------------------------------------------------------------------
// CsPreIncrementDecrementExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsPreIncrementDecrementExpression * node )
{
	addToken ( node->oper );
	if ( node->unary_expression )
		node->unary_expression->callback ( this );
}

//------------------------------------------------------------------------------
// CsCastUnaryExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsCastUnaryExpression * node )
{
	addToken ( tkLPAREN );
	if ( node->type )
		callback ( node->type );
	addToken ( tkRPAREN );
	addWS();
	if ( node->unary_expression )
		node->unary_expression->callback ( this );
}

//------------------------------------------------------------------------------
// CsAssignmentExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsAssignmentExpression * node )
{
	if ( node->lhs )
		node->lhs->callback ( this );
	addOperator ( node->oper );
	if ( node->rhs )
		node->rhs->callback ( this );
}

//------------------------------------------------------------------------------
// CsConstantExpression
//------------------------------------------------------------------------------

//#if CS_TRANSFORMATION_PARSER
//void CsFormatter::callback ( CsConstantExpression * node )
//{
//
//}
//#endif // CS_TRANSFORMATION_PARSER

//------------------------------------------------------------------------------
// CsExpressionList
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsExpressionList * node )
{
	single_list1<CsExpression*>::iterator it
		= node->list.begin();
	single_list1<CsExpression*>::iterator it_end
		= node->list.end();
	bool bFirst = true;
	while ( it != it_end )
	{
		if ( !bFirst )
			addComma();
		if ( *it )
			(*it)->callback ( this );
		bFirst = false;
		++it;
	}
}

//------------------------------------------------------------------------------
// CsNamedArgument
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsNamedArgument * node )
{
	addToken ( node->identifier );
	addOperator ( tkASSIGN );
	if ( node->expression )
		node->expression->callback ( this );
}

//------------------------------------------------------------------------------
// CsAttribute
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsAttribute * node )
{
	callback ( node->compound_identifier );
	if ( node->positional_argument_list
		|| node->named_argument_list )
	{
		addToken ( tkLPAREN );
	}
	bool bFirst = true;
	if ( node->positional_argument_list )
	{
		single_list1<CsExpression*>::iterator it
			= node->positional_argument_list->begin();
		single_list1<CsExpression*>::iterator it_end
			= node->positional_argument_list->end();
		while ( it != it_end )
		{
			if ( !bFirst )
				addComma();
			if ( *it )
				(*it)->callback ( this );
			bFirst = false;
			++it;
		}
	}
	if ( node->named_argument_list )
	{
		single_list1<CsNamedArgument*>::iterator it
			= node->named_argument_list->begin();
		single_list1<CsNamedArgument*>::iterator it_end
			= node->named_argument_list->end();
		while ( it != it_end )
		{
			if ( !bFirst )
				addComma();
			if ( (*it) )
				(*it)->callback ( this );
			bFirst = false;
			++it;
		}
	}
	if ( node->positional_argument_list
		|| node->named_argument_list )
	{
		addToken ( tkRPAREN );
	}
}

//------------------------------------------------------------------------------
// CsAttributeSection
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsAttributeSection * node )
{
	addToken ( tkLBRACKET );
	switch ( node->specifier )
	{
		case asNONE:
			break;
		case asASSEMBLY:
			addToken ( PHashString ( (const Char*)"assembly" ) );
			goto colon;
		case asEVENT:
			addToken ( PHashString ( (const Char*)"event" ) );
			goto colon;
		case asFIELD:
			addToken ( PHashString ( (const Char*)"field" ) );
			goto colon;
		case asMETHOD:
			addToken ( PHashString ( (const Char*)"method" ) );
			goto colon;
		case asPARAM:
			addToken ( PHashString ( (const Char*)"param" ) );
			goto colon;
		case asPROPERTY:
			addToken ( PHashString ( (const Char*)"property" ) );
			goto colon;
		case asRETURN:
			addToken ( PHashString ( (const Char*)"return" ) );
			goto colon;
		case asTYPE:
			addToken ( PHashString ( (const Char*)"type" ) );
			goto colon;
		default:
			ASSERT(0);
			break;
		colon:
			addColon();
			addWS();
			break;
	}
	single_list1<CsAttribute*>::iterator it
		= node->attribute_list.begin();
	single_list1<CsAttribute*>::iterator it_end
		= node->attribute_list.end();
	bool bFirst = true;
	while ( it != it_end )
	{
		if ( !bFirst )
			addComma();
		if ( *it )
			callback ( *it );
		bFirst = false;
		++it;
	}
	addToken ( tkRBRACKET );
	addNewline();
}

//------------------------------------------------------------------------------
// CsAttributes
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsAttributes * node )
{
	single_list1<CsAttributeSection*>::iterator it
		= node->sections.begin();
	single_list1<CsAttributeSection*>::iterator it_end
		= node->sections.end();
	while ( it != it_end )
	{
		if ( *it )
			callback ( *it );
		++it;
	}
}

//------------------------------------------------------------------------------
// CsTypeBase
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsTypeBase * node )
{
	addWS();
	addToken ( tkCOLON );
	single_list1<CsTypeRef*>::iterator it
		= node->base_list.begin();
	single_list1<CsTypeRef*>::iterator it_end
		= node->base_list.end();
	bool bFirst = true;
	while ( it != it_end )
	{
		if ( !bFirst )
			addComma();
		if ( *it )
			callback ( *it );
		bFirst = false;
		++it;
	}
}

//------------------------------------------------------------------------------
// CsModifiers
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsModifiers & node )
{
	int flag = 1;
	while ( flag <= mUNSAFE )
	{
		if ( node.flags & flag )
		{
			switch ( flag )
			{
				case mINTERNAL:
					addKeyword ( tkINTERNAL );
					break;
				case mPRIVATE:
					addKeyword ( tkPRIVATE );
					break;
				case mPROTECTED:
					addKeyword ( tkPROTECTED );
					break;
				case mPUBLIC:
					addKeyword ( tkPUBLIC );
					break;
				case mABSTRACT:
					addKeyword ( tkABSTRACT );
					break;
				case mEXTERN:
					addKeyword ( tkEXTERN );
					break;
				case mNEW:
					addKeyword ( tkNEW );
					break;
				case mOVERRIDE:
					addKeyword ( tkOVERRIDE );
					break;
				case mREADONLY:
					addKeyword ( tkREADONLY );
					break;
				case mSEALED:
					addKeyword ( tkSEALED );
					break;
				case mSTATIC:
					addKeyword ( tkSTATIC );
					break;
				case mVIRTUAL:
					addKeyword ( tkVIRTUAL );
					break;
				case mVOLATILE:
					addKeyword ( tkVOLATILE );
					break;
				case mUNSAFE:
					addKeyword ( tkUNSAFE );
					break;
				default:
					ASSERT(0);
			}
		}
		flag <<= 1;
	}
	//single_list1<CsTokenIndex>::node * next
	//	= & modifier_tokens.first;
	//while ( next )
	//{
	//	//tokenCollector.addToken ( next->value_ );
	//	next = next->next_;
	//}
}

//------------------------------------------------------------------------------
// CsClass
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsClass * node )
{
	addNewline2();
	if ( node->attributes )
		callback ( node->attributes );
	callback ( node->modifiers );
	addKeyword ( tkCLASS );
	addToken ( node->name );
	if ( node->base )
		callback ( node->base );
	addNewline();
	addToken ( tkLCURLY );
	FastDeque<CsNode*>::forward_iterator it
		= node->member_declarations.begin();
	FastDeque<CsNode*>::forward_iterator it_end
		= node->member_declarations.end();
	addIncTab();
	addNewlineKill1();
	for ( ; it != it_end; ++it )
	{
		if ( *it )
		{
			addNewline();
			(*it)->callback ( this );
			addNewline();
		}
	}
	addNewlineKill1();
	addDecTab();
	addToken ( tkRCURLY );
	addNewline2();
}

//------------------------------------------------------------------------------
// CsStruct
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsStruct * node )
{
	addNewline2();
	if ( node->attributes )
		callback ( node->attributes );
	callback ( node->modifiers );
	addKeyword ( tkSTRUCT );
	addToken ( node->name );
	if ( node->base )
		callback ( node->base );
	addNewline();
	addToken ( tkLCURLY );
	addIncTab();
	FastDeque<CsNode*>::forward_iterator it
		= node->member_declarations.begin();
	FastDeque<CsNode*>::forward_iterator it_end
		= node->member_declarations.end();
	for ( ; it != it_end; ++it )
	{
		if ( *it )
			(*it)->callback ( this );
	}
	addDecTab();
	addNewlineKill1();
	addToken ( tkRCURLY );
	addNewline2();
}

//------------------------------------------------------------------------------
// CsInterface
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsInterface * node )
{
	addNewline2();
	if ( node->attributes )
		callback ( node->attributes );
	callback ( node->modifiers );
	addKeyword ( tkINTERFACE );
	addToken ( node->name );
	if ( node->base )
		callback ( node->base );
	addNewline();
	addToken ( tkLCURLY );
	addIncTab();
	FastDeque<CsNode*>::forward_iterator it
		= node->member_declarations.begin();
	FastDeque<CsNode*>::forward_iterator it_end
		= node->member_declarations.end();
	for ( ; it != it_end; ++it )
	{
		addNewlineKill1();
		if ( *it )
			(*it)->callback ( this );
		addNewlineKill1();
	}
	addDecTab();
	addNewlineKill1();
	addToken ( tkRCURLY );
	addNewline2();
}

//------------------------------------------------------------------------------
// CsEnumMember
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsEnumMember * node )
{
	if ( node->attributes )
		callback ( node->attributes );
	addToken ( node->name );
	if ( node->expression )
	{
		addOperator ( tkASSIGN );
		node->expression->callback ( this );
	}
	addComma();
	addNewline();
}

//------------------------------------------------------------------------------
// CsEnum
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsEnum * node )
{
	addNewline2();
	if ( node->attributes )
		callback ( node->attributes );
	callback ( node->modifiers );
	addKeyword ( tkENUM );
	addToken ( node->name );
	if ( node->base != tkEOF )
	{
		addWS();
		addToken ( tkCOLON );
		addWS();
		addToken ( node->base );
	}
	addNewline();
	addToken ( tkLCURLY );
	addNewline();
	addIncTab();
	FastArray<CsEnumMember*>::iterator it = node->members.begin();
	FastArray<CsEnumMember*>::iterator it_end = node->members.end();
	for ( ; it != it_end; ++it )
	{
		if ( *it )
			callback ( *it );
	}
	addNewline();
	addDecTab();
	addToken ( tkRCURLY );
	addNewline2();
}

//------------------------------------------------------------------------------
// CsFormalParameter
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsFormalParameter * node )
{
	if ( node->attributes )
		callback ( node->attributes );
	if ( node->modifier != tkEOF )
		addKeyword ( node->modifier );
	if ( node->type )
		callback ( node->type );
	addToken ( node->name );
}

//------------------------------------------------------------------------------
// CsFormalParameterList
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsFormalParameterList * node )
{
	single_list1<CsFormalParameter*>::iterator it
		= node->parameters.begin();
	single_list1<CsFormalParameter*>::iterator it_end
		= node->parameters.end();
	bool bFirst = true;
	while ( it != it_end )
	{
		if ( !bFirst )
			addComma();
		if ( *it )
			callback ( *it );
		bFirst = false;
		++it;
	}
}

//------------------------------------------------------------------------------
// CsDelegate
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsDelegate * node )
{
	if ( node->attributes )
		callback ( node->attributes );
	callback ( node->modifiers );
	addToken ( tkDELEGATE );
	if ( node->type )
		callback ( node->type );
	addToken ( node->name );
	addWS();
	addToken ( tkLPAREN );
	addWS();
	if ( node->parameters )
		callback ( node->parameters );
	addWS();
	addToken ( tkRPAREN );
	addSemicolon();
}

//------------------------------------------------------------------------------
// CsNamespace
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsNamespace * node )
{
	if ( node->attributes )
		callback ( node->attributes );
	callback ( node->modifiers );
	addKeyword ( tkNAMESPACE );
	callback ( node->compound_identifier );
	addNewline();
	addToken ( tkLCURLY );
	addNewlineKill1();
	addNewline();
	addIncTab();
	{
		FastArray<CsUsingDirective*>::iterator it = node->using_directives.begin();
		FastArray<CsUsingDirective*>::iterator it_end = node->using_directives.end();;
		for ( ; it != it_end; ++it )
		{
			if ( *it )
			{
				addNewline();
				(*it)->callback ( this );
				addNewline();
			}
		}
	}
	{
		FastArray<CsNode*>::iterator it = node->member_declarations.begin();
		FastArray<CsNode*>::iterator it_end = node->member_declarations.end();;
		for ( ; it != it_end; ++it )
		{
			if ( *it )
			{
				addNewline2();
				(*it)->callback ( this );
				addNewline2();
			}
		}
	}
	addDecTab();
	addNewline();
	addNewlineKill1();
	addToken ( tkRCURLY );
	addNewline();
}

//------------------------------------------------------------------------------
// CsArgument
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsArgument * node )
{
	if ( node->modifier != tkEOF )
		addKeyword ( node->modifier );
	if ( node->expression )
		node->expression->callback ( this );
}

//------------------------------------------------------------------------------
// CsArgumentList
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsArgumentList * node )
{
	single_list1<CsArgument*>::iterator it
		= node->list.begin();
	single_list1<CsArgument*>::iterator it_end
		= node->list.end();
	bool bFirst = true;
	while ( it != it_end )
	{
		if ( !bFirst )
			addComma();
		if ( *it )
			callback ( *it );
		bFirst = false;
		++it;
	}
}

//------------------------------------------------------------------------------
// CsLabel
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsLabel * node )
{
	addToken ( node->identifier );
	addColon();
	addNewline();
}

//------------------------------------------------------------------------------
// CsLabeledStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsLabeledStatement * node )
{
	if ( node->labels )
	{
		addDecTab();
		single_list1<CsLabel*>::iterator it
			= node->labels->begin();
		single_list1<CsLabel*>::iterator it_end
			= node->labels->end();
		while ( it != it_end )
		{
			(*it)->callback ( this );
			++it;
		}
		addIncTab();
	}
	if ( node->statement )
		node->statement->callback ( this );
}

//------------------------------------------------------------------------------
// CsEmptyStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsEmptyStatement * node )
{
	addSemicolon();
}

//------------------------------------------------------------------------------
// CsExpressionStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsExpressionStatement * node )
{
	if ( node->expression )
		node->expression->callback ( this );
	addSemicolon();
}

//------------------------------------------------------------------------------
// CsDeclarationStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsDeclarationStatement * node )
{
	if ( node->node )
		node->node->callback ( this );
	addSemicolon();
}

//------------------------------------------------------------------------------
// CsLocalConstantDeclarator
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsLocalConstantDeclarator * node )
{
	if ( node->name )
		addToken ( node->name );
	if ( node->expression )
	{
		addOperator ( tkASSIGN );
		node->expression->callback ( this );
	}
}

//------------------------------------------------------------------------------
// CsLocalConstantDeclaration
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsLocalConstantDeclaration * node )
{
	addKeyword ( tkCONST );
	if ( node->type )
		callback ( node->type );
	single_list1<CsLocalConstantDeclarator*>::iterator it
		= node->declarators.begin();
	single_list1<CsLocalConstantDeclarator*>::iterator it_end
		= node->declarators.end();
	bool bFirst = true;
	while ( it != it_end )
	{
		if ( !bFirst )
			addComma();
		if ( *it )
			callback ( *it );
		bFirst = false;
		++it;
	}
}

//------------------------------------------------------------------------------
// CsIfStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsIfStatement * node )
{
	addKeyword ( tkIF );
	addToken ( tkLPAREN );
	addWS();
	if ( node->condition )
		node->condition->callback ( this );
	addWS();
	addToken ( tkRPAREN );
	if ( node->if_statement )
	{
		addNewline();
		bool bBlock = node->if_statement->type == n_block;
		if ( !bBlock )
			addIncTab();
		node->if_statement->callback ( this );
		if ( !bBlock )
			addDecTab();
		addNewline();
	}
	if ( node->else_statement )
	{
		addToken ( tkELSE );
		bool bIfStatement = node->else_statement->type == n_if_statement;
		if ( !bIfStatement )
			addNewline();
		bool bBlock = node->else_statement->type == n_block;
		if ( !bBlock && !bIfStatement )
			addIncTab();
		node->else_statement->callback ( this );
		if ( !bBlock && !bIfStatement )
			addDecTab();
		addNewline();
	}
}

//------------------------------------------------------------------------------
// CsSwitchLabel
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsSwitchLabel * node )
{
	addNewline();
	if ( node->expression )
	{
		addKeyword ( tkCASE );
		if ( node->expression )
			node->expression->callback ( this );
		addColon();
	}
	else
	{
		addKeyword ( tkDEFAULT );
		addColon();
	}
	addNewline();
}

//------------------------------------------------------------------------------
// CsSwitchSection
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsSwitchSection * node )
{
	if ( node->labels )
	{
		single_list1<CsSwitchLabel*>::iterator it
			= node->labels->begin();
		single_list1<CsSwitchLabel*>::iterator it_end
			= node->labels->end();
		while ( it != it_end )
		{
			if ( *it )
				callback ( *it );
			++it;
		}
	}
	addIncTab();
	FastArray<CsStatement*>::iterator it = node->statements.begin();
	FastArray<CsStatement*>::iterator it_end = node->statements.end();
	for ( ; it != it_end; ++it )
	{
		if ( *it )
		{
			addNewline();
			(*it)->callback ( this );
			addNewline();
		}
	}
	addDecTab();
}

//------------------------------------------------------------------------------
// CsSwitchStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsSwitchStatement * node )
{
	addKeyword ( tkSWITCH );
	addToken ( tkLPAREN );
	addWS();
	if ( node->expression )
		node->expression->callback ( this );
	addWS();
	addToken ( tkRPAREN );
	addNewline();
	addToken ( tkLCURLY );
	addNewline();
	addIncTab();
	FastArray<CsSwitchSection*>::iterator it = node->sections.begin();
	FastArray<CsSwitchSection*>::iterator it_end = node->sections.end();
	for ( ; it != it_end; ++it )
	{
		if ( *it )
			callback ( *it );
	}
	addDecTab();
	addNewline();
	addToken ( tkRCURLY );
}

//------------------------------------------------------------------------------
// CsWhileStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsWhileStatement * node )
{
	addKeyword ( tkWHILE );
	addWS();
	addToken( tkLPAREN );
	addWS();
	if ( node->condition )
		node->condition->callback ( this );
	addWS();
	addToken( tkRPAREN );
	addNewline();
	bool bBlock = node->statement->type == n_block;
	if ( node->statement )
	{
		addNewline();
		if ( !bBlock )
			addIncTab();
		node->statement->callback ( this );
		if ( !bBlock )
			addDecTab();
	}
}

//------------------------------------------------------------------------------
// CsDoWhileStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsDoWhileStatement * node )
{
	addKeyword ( tkDO );
	addNewline();
	bool bBlock = node->statement->type == n_block;
	if ( node->statement )
	{
		if ( !bBlock )
			addIncTab();
		node->statement->callback ( this );
		if ( !bBlock )
			addDecTab();
	}
	addNewline();
	addKeyword ( tkWHILE );
	addToken ( tkLPAREN );
	if ( node->condition )
	{
		addWS();
		node->condition->callback ( this );
		addWS();
	}
	addToken ( tkRPAREN );
	addSemicolon();
}

//------------------------------------------------------------------------------
// CsLocalVariableDeclarator
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsLocalVariableDeclarator * node )
{
	addToken ( node->name );
	if ( node->expression )
	{
		addOperator ( tkASSIGN );
		node->expression->callback ( this );
	}
}

//------------------------------------------------------------------------------
// CsLocalVariableDeclaration
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsLocalVariableDeclaration * node )
{
	if ( node->type )
		node->type->callback ( this );
	single_list1<CsLocalVariableDeclarator*>::iterator it
		= node->declarators.begin();
	single_list1<CsLocalVariableDeclarator*>::iterator it_end
		= node->declarators.end();
	bool bFirst = true;
	while ( it != it_end )
	{
		if ( !bFirst )
			addComma();
		if ( *it )
			callback ( *it );
		bFirst = false;
		++it;
	}
}

//------------------------------------------------------------------------------
// CsStatementExpressionList
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsStatementExpressionList * node )
{
	single_list1<CsExpression*>::iterator it
		= node->expressions.begin();
	single_list1<CsExpression*>::iterator it_end
		= node->expressions.end();
	bool bFirst = true;
	while ( it != it_end )
	{
		if ( !bFirst )
			addComma();
		//tokenCollector.addToken ( next->value_.separator_token );
		if ( *it )
			(*it)->callback ( this );
		bFirst = false;
		++it;
	}
}

//------------------------------------------------------------------------------
// CsForStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsForStatement * node )
{
	addKeyword ( tkFOR );
	addToken ( tkLPAREN );
	addWS();
	if ( node->initializer )
	{
		node->initializer->callback ( this );
		addSemicolon();
	}
	else
	{
		addToken ( tkSEMICOLON );
	}
	addWS();
	if ( node->condition )
		node->condition->callback ( this );
	addSemicolon();
	addWS();
	if ( node->iterator )
		node->iterator->callback ( this );
	addWS();
	addToken ( tkRPAREN );
	addNewline();
	bool bBlock = node->statement->type == n_block;
	if ( node->statement )
	{
		if ( !bBlock )
			addIncTab();
		node->statement->callback ( this );
		if ( !bBlock )
			addDecTab();
	}
}

//------------------------------------------------------------------------------
// CsForeachStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsForeachStatement * node )
{
	addKeyword ( tkFOREACH );
	addToken ( tkLPAREN );
	if ( node->type )
		callback ( node->type );
	addToken ( node->identifier );
	addKeyword ( tkIN );
	if ( node->expression )
		node->expression->callback ( this );
	addWS();
	addToken ( tkRPAREN );
	addNewline();
	bool bBlock = node->statement->type == n_block;
	if ( node->statement )
	{
		if ( !bBlock )
			addIncTab();
		node->statement->callback ( this );
		if ( !bBlock )
			addDecTab();
	}
}

//------------------------------------------------------------------------------
// CsBreakStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsBreakStatement * node )
{
	addKeyword ( tkBREAK );
	addSemicolon();
}

//------------------------------------------------------------------------------
// CsContinueStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsContinueStatement * node )
{
	addKeyword ( tkCONTINUE );
	addSemicolon();
}

//------------------------------------------------------------------------------
// CsGotoStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsGotoStatement * node )
{
	addKeyword ( tkGOTO );
	switch ( node->label_type )
	{
		case tkIDENTIFIER:
			addToken ( node->label );
			break;
		case tkCASE:
			addKeyword ( tkCASE );
			if ( node->case_expression )
				node->case_expression->callback ( this );
			break;
		case tkDEFAULT:
			addToken ( tkDEFAULT );
			break;
	}
	addSemicolon();
}

//------------------------------------------------------------------------------
// CsReturnStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsReturnStatement * node )
{
	addKeyword ( tkRETURN );
	if ( node->expression )
		node->expression->callback ( this );
	addSemicolon();
}

//------------------------------------------------------------------------------
// CsThrowStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsThrowStatement * node )
{
	addKeyword ( tkTHROW );
	if ( node->expression )
		node->expression->callback ( this );
	addSemicolon();
}

//------------------------------------------------------------------------------
// CsCatchClause
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsCatchClause * node )
{
	addKeyword ( tkCATCH );
	if ( node->type )
	{
		addToken ( tkLPAREN );
		addWS();
		callback ( node->type );
		addWS();
		addToken ( node->identifier );
		addWS();
		addToken ( tkRPAREN );
	}
	addNewline();
	if ( node->block )
		callback ( node->block );
}

//------------------------------------------------------------------------------
// CsTryStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsTryStatement * node )
{
	addKeyword ( tkTRY );
	addNewline();
	if ( node->try_block )
		callback ( node->try_block );
	single_list1<CsCatchClause*>::iterator it
		= node->catch_clauses.begin();
	single_list1<CsCatchClause*>::iterator it_end
		= node->catch_clauses.end();
	while ( it != it_end )
	{
		addNewline();
		if ( *it )
			callback ( *it );
		++it;
	}
	if ( node->finally_block )
	{
		addNewline();
		addKeyword ( tkFINALLY );
		addNewline();
		callback ( node->finally_block );
	}
}

//------------------------------------------------------------------------------
// CsCheckedStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsCheckedStatement * node )
{
	addKeyword ( tkCHECKED );
	addNewline();
	if ( node->block )
		callback ( node->block );
}

//------------------------------------------------------------------------------
// CsUncheckedStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsUncheckedStatement * node )
{
	addKeyword ( tkUNCHECKED );
	addNewline();
	if ( node->block )
		callback ( node->block );
}

//------------------------------------------------------------------------------
// CsLockStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsLockStatement * node )
{
	addKeyword ( tkLOCK );
	addToken ( tkLPAREN );
	addWS();
	if ( node->expression )
		node->expression->callback ( this );
	addWS();
	addToken ( tkRPAREN );
	addNewline();
	bool bBlock = node->statement->type == n_block;
	if ( node->statement )
	{
		if ( !bBlock )
			addIncTab();
		node->statement->callback ( this );
		if ( !bBlock )
			addDecTab();
	}
}

//------------------------------------------------------------------------------
// CsUsingStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsUsingStatement * node )
{
	addKeyword ( tkUSING );
	addToken ( tkLPAREN );
	addWS();
	if ( node->resource )
		node->resource->callback ( this );
	addWS();
	addToken ( tkRPAREN );
	addNewline();
	bool bBlock = node->statement->type == n_block;
	if ( node->statement )
	{
		if ( !bBlock )
			addIncTab();
		node->statement->callback ( this );
		if ( !bBlock )
			addDecTab();
	}
}

//------------------------------------------------------------------------------
// CsUnsafeStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsUnsafeStatement * node )
{
	addKeyword ( tkUNSAFE );
	addNewline();
	if ( node->block )
		callback ( node->block );
}

//------------------------------------------------------------------------------
// CsFixedPointDeclarator
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsFixedPointDeclarator * node )
{
	addToken ( node->identifier );
	addWS();
	addOperator ( tkASSIGN );
	if ( node->expression )
		node->expression->callback ( this );
}

//------------------------------------------------------------------------------
// CsFixedStatement
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsFixedStatement * node )
{
	addKeyword ( tkFIXED );
	addWS();
	addToken ( tkLPAREN );
	addWS();
	if ( node->type )
		callback ( node->type );
	single_list1<CsFixedPointDeclarator*>::iterator it
		= node->declarators.begin();
	single_list1<CsFixedPointDeclarator*>::iterator it_end
		= node->declarators.end();
	bool bFirst = true;
	while ( it != it_end )
	{
		if ( !bFirst )
			addComma();
		if ( *it )
			callback ( *it );
		bFirst = false;
		++it;
	}
	addWS();
	addToken ( tkRPAREN );
	addNewline();
	bool bBlock = node->statement->type == n_block;
	if ( node->statement )
	{
		if ( !bBlock )
			addIncTab();
		node->statement->callback ( this );
		if ( !bBlock )
			addDecTab();
	}
}

//------------------------------------------------------------------------------
// CsBlock
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsBlock * node )
{
	addToken ( tkLCURLY );
	FastArray<CsStatement*>::iterator it = node->statements.begin();
	FastArray<CsStatement*>::iterator it_end = node->statements.end();
	addIncTab();
	for ( ; it != it_end; ++it )
	{
		addNewline();
		if ( *it )
			(*it)->callback ( this );
		addNewline();
	}
	addDecTab();
	addToken ( tkRCURLY );
}

//------------------------------------------------------------------------------
// CsStackAllocExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsStackAllocExpression * node )
{
	addKeyword ( tkSTACKALLOC );
	addWS();
	if ( node->type )
		callback ( node->type );
	addToken ( tkLBRACKET );
	addWS();
	if ( node->expression )
		node->expression->callback ( this );
	addWS();
	addToken ( tkRBRACKET );
}

//------------------------------------------------------------------------------
// CsConstructorDeclaration
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsConstructorDeclaration * node )
{
	if ( node->definition
		|| node->basethis != tkEOF )
	{
		addNewline2();
	}
	if ( node->attributes )
		callback ( node->attributes );
	callback ( node->modifiers );
	addToken ( node->name );
	addWS();
	addToken ( tkLPAREN );
	addWS();
	if ( node->parameters )
		callback ( node->parameters );
	addWS();
	addToken ( tkRPAREN );
	addNewline();
	if ( node->basethis != tkEOF )
	{
		addIncTab();
		addToken ( tkCOLON );
		addWS();
		addKeyword ( node->basethis );
		addToken ( tkLPAREN );
		if ( node->argument_list )
		{
			addWS();
			callback ( node->argument_list );
			addWS();
		}
		addToken ( tkRPAREN );
		addDecTab();
		addNewline();
	}
	if ( node->definition )
	{
		addNewline();
		callback ( node->definition );
	}
	else
	{
		addSemicolon();
	}
	if ( node->definition
		|| node->basethis != tkEOF )
	{
		addNewline2();
	}
}

//------------------------------------------------------------------------------
// CsConstantDeclarator
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsConstantDeclarator * node )
{
	addToken ( node->name );
	if ( node->expression )
	{
		addWS();
		addOperator ( tkASSIGN );
		addWS();
		if ( node->expression )
			node->expression->callback ( this );
	}
}

//------------------------------------------------------------------------------
// CsConstantDeclaration
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsConstantDeclaration * node )
{
  if ( node->attributes )
		addNewline2();
	else
		addNewline();
	if ( node->attributes )
		callback ( node->attributes );
	callback ( node->modifiers );
	addKeyword ( tkCONST );
	addWS();
	if ( node->type )
		callback ( node->type );
	addWS();
	single_list1<CsConstantDeclarator*>::iterator it
		= node->declarators.begin();
	single_list1<CsConstantDeclarator*>::iterator it_end
		= node->declarators.end();
	bool bFirst = true;
	while ( it != it_end )
	{
		if ( !bFirst )
			addComma();
		if ( *it )
			callback ( *it );
		bFirst = false;
		++it;
	}
	addSemicolon();
  if ( node->attributes )
		addNewline2();
	else
		addNewline();
}

//------------------------------------------------------------------------------
// CsVariableDeclarator
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsVariableDeclarator * node )
{
	addToken ( node->name );
	if ( node->expression )
	{
		addWS();
		addOperator ( tkASSIGN );
		addWS();
		if ( node->expression )
			node->expression->callback ( this );
	}
}

//------------------------------------------------------------------------------
// CsVariableDeclaration
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsVariableDeclaration * node )
{
	if ( node->attributes )
		addNewline2();
	else
		addNewline();
	if ( node->attributes )
		callback ( node->attributes );
	callback ( node->modifiers );
	addWS();
	if ( node->type )
		callback ( node->type );
	addWS();
	single_list1<CsVariableDeclarator*>::iterator it
		= node->declarators.begin();
	single_list1<CsVariableDeclarator*>::iterator it_end
		= node->declarators.end();
	bool bFirst = true;
	while ( it != it_end )
	{
		if ( !bFirst )
			addComma();
		if ( *it )
			callback ( *it );
		bFirst = false;
		++it;
	}
	addSemicolon();
	if ( node->attributes )
		addNewline2();
	else
		addNewline();
}

//------------------------------------------------------------------------------
// CsArrayInitializer
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsArrayInitializer * node )
{
	addToken ( tkLCURLY );
	FastDeque<CsExpression*>::forward_iterator it = node->expressions.begin();
	FastDeque<CsExpression*>::forward_iterator it_end = node->expressions.end();
	bool bFirst = true;
	for ( ; it != it_end; ++it )
	{
		if ( !bFirst )
			addComma();
		if ( (*it) )
		{
			addWS();
			(*it)->callback ( this );
			addWS();
		}
		bFirst = false;
	}
	addToken ( tkRCURLY );
}

//------------------------------------------------------------------------------
// CsMethod
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsMethod * node )
{
	if ( node->definition
		|| node->attributes )
	{
		addNewline2();
	}
	else
	{
		addNewline();
	}
	if ( node->attributes )
		callback ( node->attributes );
	callback ( node->modifiers );
	if ( node->return_type )
		callback ( node->return_type );
	callback ( node->compound_identifier );
	addWS();
	addToken ( tkLPAREN );
	addWS();
	if ( node->parameters )
		callback ( node->parameters );
	addWS();
	addToken ( tkRPAREN );
	if ( node->definition )
	{
		addNewline();
		callback ( node->definition );
	}
	else
	{
		addSemicolon();
	}
	if ( node->definition
		|| node->attributes )
	{
		addNewline2();
	}
	else
	{
		addNewline();
	}
}

//------------------------------------------------------------------------------
// CsPropertyAccessor
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsPropertyAccessor * node )
{
	if ( node->attributes )
		callback ( node->attributes );
	if ( !node->bSet )
		addToken ( PHashString ( (const Char*)"get" ) );
	else
		addToken ( PHashString ( (const Char*)"set" ) );
	if ( node->definition )
	{
		addNewline();
		callback ( node->definition );
	}
	else
	{
		addSemicolon();
	}
	addNewline();
}

//------------------------------------------------------------------------------
// CsProperty
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsProperty * node )
{
	addNewline2();
	if ( node->attributes )
		callback ( node->attributes );
	callback ( node->modifiers );
	if ( node->type )
		callback ( node->type );
	callback ( node->compound_identifier );
	addNewline();
	addToken ( tkLCURLY );
	addNewline();
	addIncTab();
	if ( node->get )
		callback ( node->get );
	if ( node->set )
		callback ( node->set );
	addDecTab();
	addNewline();
	addToken ( tkRCURLY );
	addNewline2();
}

//------------------------------------------------------------------------------
// CsEventAccessor
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsEventAccessor * node )
{
	if ( node->attributes )
		callback ( node->attributes );
	callback ( node->modifiers );
	addWS();
	if ( node->bAdd )
		addToken ( PHashString ( (const Char*)"add" ) );
	else
		addToken ( PHashString ( (const Char*)"remove" ) );
	if ( node->definition )
	{
		addNewline();
		callback ( node->definition );
	}
	else
		addSemicolon();
	addNewline();
}

//------------------------------------------------------------------------------
// CsEventDeclarator
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsEventDeclarator * node )
{
	callback ( node->compound_identifier );
	bool bCurly = node->add || node->remove;
	if ( !node->expression )
	{
		addNewline();
		if ( bCurly )
			addToken ( tkLCURLY );
		addNewline();
		addIncTab();
		if ( node->add )
			callback ( node->add );
		if ( node->remove )
			callback ( node->remove );
		addDecTab();
		addNewline();
		if ( bCurly )
			addToken ( tkRCURLY );
	}
	else
	{
		addToken ( tkASSIGN );
		if ( node->expression )
			node->expression->callback ( this );
		addSemicolon();
	}
}

//------------------------------------------------------------------------------
// CsEvent
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsEvent * node )
{
	addNewline2();
	if ( node->attributes )
		callback ( node->attributes );
	callback ( node->modifiers );
	addKeyword ( tkEVENT );
	if ( node->type )
		callback ( node->type );
	single_list1<CsEventDeclarator*>::iterator it
		= node->declarators.begin();
	single_list1<CsEventDeclarator*>::iterator it_end
		= node->declarators.end();
	bool bSemicolon = true;
	if ( it != it_end )
	{
		if ( (*it)->add
		  || (*it)->remove )
		{
			bSemicolon = false;
		}
	}
	while ( it != it_end )
	{
		if ( (*it) )
			callback ( *it );
		++it;
	}
	if ( bSemicolon )
		addSemicolon();
	addNewline2();
}

//------------------------------------------------------------------------------
// CsIndexer
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsIndexer * node )
{
	addNewline2();
	if ( node->attributes )
		callback ( node->attributes );
	callback ( node->modifiers );
	if ( node->type )
		callback ( node->type );
	callback ( node->compound_prefix );
	if ( node->compound_prefix.front().identifier )
		addToken ( tkDOT );
	addToken ( tkTHIS );
	addWS();
	addToken ( tkLBRACKET );
	addWS();
	if ( node->parameters )
		callback ( node->parameters );
	addWS();
	addToken ( tkRBRACKET );
	bool bGetSet = node->get || node->set;
	if ( bGetSet )
	{
		addNewline();
		addToken ( tkLCURLY );
		addNewline();
		addIncTab();
		if ( node->get )
			callback ( node->get );
		if ( node->set )
			callback ( node->set );
		addDecTab();
		addNewline();
		addToken ( tkRCURLY );
	}
	else
	{
		addSemicolon();
	}
	addNewline2();
}

//------------------------------------------------------------------------------
// CsOperator
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsOperator * node )
{
	if ( node->attributes
		|| node->definition )
	{
		addNewline2();
	}
	else
	{
		addNewline();
	}
	if ( node->attributes )
		callback ( node->attributes );
	callback ( node->modifiers );
	if ( node->type )
		callback ( node->type );
	addKeyword ( tkOPERATOR );
	addOperator ( node->operation );
	addToken ( tkLPAREN );
	//if ( node->type1 )
	//{
	//	addWS();
	//	callback ( node->type1 );
	//	addToken ( node->id1 );
	//	addWS();
	//}
	if ( node->parameter1 )
	{
		addWS();
		callback ( node->parameter1 );
		addWS();
	}
	//if ( node->type2 )
	//{
	//	addComma();
	//	callback ( node->type2 );
	//	addToken ( node->id2 );
	//	addWS();
	//}
	if ( node->parameter2 )
	{
		addComma();
		addWS();
		callback ( node->parameter2 );
		addWS();
	}
	addToken ( tkRPAREN );
	if ( node->definition )
	{
		addNewline();
		callback ( node->definition );
	}
	else
	{
		addSemicolon();
	}
	if ( node->attributes
		|| node->definition )
	{
		addNewline2();
	}
	else
	{
		addNewline();
	}
}

//------------------------------------------------------------------------------
// CsConversionOperatorDeclaration
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsConversionOperatorDeclaration * node )
{
	if ( node->attributes
		|| node->definition )
	{
		addNewline2();
	}
	else
	{
		addNewline();
	}
	if ( node->attributes )
		callback ( node->attributes );
	callback ( node->modifiers );
	switch ( node->modifier )
	{
		case tkEXPLICIT:
		case tkIMPLICIT:
			addKeyword ( node->modifier );
			break;
	}
	addKeyword ( tkOPERATOR );
	if ( node->type )
		callback ( node->type );
	addToken ( tkLPAREN );
	addWS();
	//if ( node->type1 )
	//	callback ( node->type1 );
	//addToken ( node->id1 );
	if ( node->parameter1 )
		callback ( node->parameter1 );
	addWS();
	addToken ( tkRPAREN );
	if ( node->definition )
	{
		addNewline();
		callback ( node->definition );
	}
	else
	{
		addSemicolon();
	}
	if ( node->attributes
		|| node->definition )
	{
		addNewline2();
	}
	else
	{
		addNewline();
	}
}

//------------------------------------------------------------------------------
// CsDestructorDeclaration
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsDestructorDeclaration * node )
{
	if ( node->attributes )
		callback ( node->attributes );
	addToken ( tkCOMPL );
	addWSKill();
	addToken ( node->name );
	addWSKill();
	addToken ( tkLPAREN );
	addToken ( tkRPAREN );
	if ( node->definition )
	{
		addNewline();
		callback ( node->definition );
	}
	else
	{
		addSemicolon();
	}
}

//------------------------------------------------------------------------------
// CsCompilationUnit
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsCompilationUnit * node )
{
	{
		FastArray<CsUsingDirective*>::iterator it = node->using_directives.begin();
		FastArray<CsUsingDirective*>::iterator it_end = node->using_directives.end();
		for ( ; it != it_end; ++it )
		{
			if ( *it )
			{
				addNewline();
				(*it)->callback ( this );
				addNewline();
			}
		}
	}
	addNewline2();
	{
		FastArray<CsAttributes*>::iterator it = node->attributes.begin();
		FastArray<CsAttributes*>::iterator it_end = node->attributes.end();
		for ( ; it != it_end; ++it )
		{
			if ( *it )
			{
				addNewline();
				callback ( *it );
				addNewline();
			}
		}
	}
	addNewline2();
	{
		FastArray<CsNode*>::iterator it = node->declarations.begin();
		FastArray<CsNode*>::iterator it_end = node->declarations.end();
		for ( ; it != it_end; ++it )
		{
			if ( *it )
			{
				addNewline2();
				(*it)->callback ( this );
				addNewline2();
			}
		}
	}
}

//------------------------------------------------------------------------------
// CsMakeRefExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsMakeRefExpression * node )
{
	addKeyword ( tkMAKEREF );
	addToken ( tkLPAREN );
	addWS();
	if ( node->expression )
		node->expression->callback ( this );
	addWS();
	addToken ( tkRPAREN );
}

//------------------------------------------------------------------------------
// CsRefTypeExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsRefTypeExpression * node )
{
	addKeyword ( tkREFTYPE );
	addToken ( tkLPAREN );
	addWS();
	if ( node->expression )
		node->expression->callback ( this );
	addWS();
	addToken ( tkRPAREN );
}

//------------------------------------------------------------------------------
// CsRefValueExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsRefValueExpression * node )
{
	addKeyword ( tkREFVALUE );
	addToken ( tkLPAREN );
	addWS();
	if ( node->expression )
		node->expression->callback ( this );
	addComma();
	if ( node->type )
		callback ( node->type );
	addWS();
	addToken ( tkRPAREN );
}

//------------------------------------------------------------------------------
// CsArgListExpression
//------------------------------------------------------------------------------

void CsFormatter::callback ( CsArgListExpression * node )
{
	addKeyword ( tkARGLIST );
	if ( node->expression_list )
	{
		addToken ( tkLPAREN );
		callback ( node->expression_list );
		addToken ( tkRPAREN );
	}
}

//------------------------------------------------------------------------------

